

set -ex



python -m pip check
conda --help
conda build --help
conda convert --help
conda develop --help
conda inspect --help
conda inspect linkages --help
conda inspect objects --help
conda metapackage --help
conda render --help
conda skeleton --help
conda debug --help
conda-build --help
conda-convert --help
conda-develop --help
conda-inspect --help
conda-inspect linkages --help
conda-inspect objects --help
conda-metapackage --help
conda-render --help
conda-skeleton --help
conda-debug --help
exit 0
